/****************************************************************************************/
/*	Titulo del Proyecto: Demo para receptor HT6P20                                      */
/*  Descripcion: El proyecto consiste en en realizar una demo para el receptor HT6P20   */
/*  En la cual se reciba la cadena de datos enviadas por HT6P20 y mostrar los datos     */
/*  recibidos en un display. En el display se mostrara la direccion del HT6P20 y el     */
/*  estado de las entradas.                                                             */
/****************************************************************************************/

/****************************************************************************************/
/*        Autor: Ing. Ivn Csar Sierra                                                 */
/*        Fecha de inicio: 09/01/2013                																		*/
/*        Fecha de finalizacion: 11/01/2013      																				*/
/*				Compilador: CCS																																*/
/*        Version: 1.0                                        													*/
/*        Actualizaciones:                                															*/
/****************************************************************************************/

/****************************************************************************************/
/* 												INFORMACION REFERENTE AL HT6P20X															*/
/****************************************************************************************/
/* El HT6P20 tiene el siguiente string: Address+Data+EndCode														*/
/* La longitud del dato varia segun la version. En el caso de la B es de 22bits y en el */
/* caso de la D es de 20bits. 																													*/
/* La direccion y el dato se envian desde el LSB al MSB por lo tanto una vez parseado   */
/* el string se tiene que invertir el orden de los bits.																*/
/* Por ejemplo se recibe el siguiente string:																						*/
/* 		0001010101111100001101110101																											*/
/*																																											*/
/* El parseo para el HT6P20B seria: 																										*/
/* 		0001010101111100001101 11 0101 																										*/
/* 		donde: 	0001010101111100001101: es la direccion																		*/
/* 						11: es el dato																														*/
/* 						0101: es el EndCode																												*/		
/* Para obtener la direccion y el dato hay que invertir los bits, por lo tanto la 			*/
/* direccion real es:																																		*/
/* 		01 1100 0011 1110 1010 1000 = 0x2C3EA8																						*/
/* y el dato real es:																																		*/
/*		11 = 0x3																																					*/
/* 																																											*/
/* El parseo para el HT6P20D seria: 																										*/
/* 		00010101011111000011 0111 0101 																										*/
/* 		donde: 	00010101011111000011: es la direccion																			*/
/* 						0111: es el dato																													*/
/* 						0101: es el EndCode																												*/
/* Para obtener la direccion y el dato hay que invertir los bits, por lo tanto la 			*/
/* direccion real es:																																		*/
/* 		1100 0011 1110 1010 1000 = 0xC3EA8																								*/
/* y el dato real es:																																		*/
/*		1110 = 0xE																																				*/
/*																																											*/
/****************************************************************************************/

// Definiciones para el compilador -------------------------------------------------------
#include <16F628A.h>	
#fuses INTRC_IO, NOWDT, BROWNOUT, NOPROTECT, MCLR
#use delay (internal = 4M)
#use fast_io (B)

// Includes ------------------------------------------------------------------------------
#include "LcdDriver.c"

// Definiciones --------------------------------------------------------------------------

// longitud del paquete de datos para el HT6P20x2 en bits
#define HT6P20X2_LEN_PKG				28

// data que indica el final de un paquete
#define END_CODE								0b00000101

// longitud de la direccion segun el TX
#define HT6P20B_LEN_ADDRESS			(long long)22
#define HT6P20D_LEN_ADDRESS			(long long)20

// Valor maximo del error que se puede tener en el ancho del pulso lambda
#define LAMBDA_ERROR						(int)40	// TMR1 se incrementa cada 1uSeg => 40uSeg

// Lcd Pin Backlight
#define LCD_Backlight 					PIN_B1

// Para el buzzer
#define Buzzer 									PIN_B0

// para la entrada de RX
#define RX_IN 									PIN_B4

// para seleccionar el modelo de encoder (HT6P20B o HT6P20D)
#define RX_MODEL 								PIN_B2

// macros --------------------------------------------------------------------------------
#define Lcd_Backlight_On 				output_high (LCD_Backlight)
#define Lcd_Backlight_Off				output_low (LCD_Backlight)
#define Buzzer_On								output_high (Buzzer)
#define Buzzer_Off							output_low (Buzzer)

// Constantes globales -------------------------------------------------------------------
enum {
	HT6P20B = 0,
	HT6P20D
};

enum {
	RX_GET_OK = 0,
	RX_GET_ERROR
};

enum {
	RX_IDLE = 0,
	RX_GET_LAMBDA,
	RX_GET_PACKAGE,
	RX_GET_PACKAGE_OK,
	RX_GET_PACKAGE_ERROR
};

// Variables Globales --------------------------------------------------------------------
typedef struct {
	unsigned char Model;
	unsigned long long Package;
	unsigned long long Address;
	unsigned char strAddress[9];
	unsigned char Data;
	unsigned char strData[3];
	unsigned char State;
	unsigned char Result;
	unsigned char CounterBits;
} RXInfo;

RXInfo RX;


// Funciones -----------------------------------------------------------------------------

// funcion para convertir un byte en un char
char ByteToChar (unsigned char byte_data) {
		
	if (byte_data < 0x0A){
		return (char)(byte_data += 0x30); 
	}
	else {
		switch(byte_data) {
			case 0x0A:
				return (char)(byte_data = 'A');
			break;
			
			case 0x0B:
				return (char)(byte_data = 'B');
			break;
			
			case 0x0C:
				return (char)(byte_data = 'C');
			break;
			
			case 0x0D:
				return (char)(byte_data = 'D');
			break;
			
			case 0x0E:
				return (char)(byte_data = 'E');
			break;
			
			case 0x0F:
				return (char)(byte_data = 'F');
			break;
		}
	}
}

// funcion para cnvertir un Long en un string en hexdecimal
void LongToStrHex (unsigned long long long_data, char *str_hex) {
	
	unsigned char i = 0, j = 0, nibble_high = 0x00, nibble_low = 0x00; 
	char *ptr_char;
	
	ptr_char = (char*)&long_data;
	
	for (i = 0; i < 4; i++) {
		nibble_high = *ptr_char;
		nibble_high &= 0xF0;
		nibble_high >>= 4;
		nibble_low = *ptr_char;
		nibble_low &= 0x0F;
		str_hex[7 - j] = ByteToChar (nibble_low);
		j++;
		str_hex[7 - j] = ByteToChar (nibble_high);
		j++;
		ptr_char++;
	}
	str_hex[j] = 0x00;
}

// funcion para convertir un byte en un string en hexadecimal
void ByteToStrHex (unsigned char chr_data, char *str_hex) {
	
	unsigned char nibble_high = 0x00, nibble_low = 0x00; 
	char *ptr_char;
	
	ptr_char = &chr_data;
	
	nibble_high = *ptr_char;
	nibble_high &= 0xF0;
	nibble_high >>= 4;
	nibble_low = *ptr_char;
	nibble_low &= 0x0F;
	str_hex[0] = ByteToChar (nibble_high);
	str_hex[1] = ByteToChar (nibble_low);
	str_hex[2] = 0x00;
}
// funcion para el control de interrupciones----------------------------------------------

// interrupcion por cambio en el puerto RB
#INT_RB
void interrupt_Ext () {
	
	static unsigned int Lambdatime = 0;							// almacena el valor del ancho del pulso lambda
	static unsigned int PulseHighTime = 0;					// almacena el valor del ancho del pulso en alto
	static unsigned int PulseLowTime = 0;						// almacena el valor del ancho del pulso en bajo
	
	// maquina de estados para la lectura de los bits
  switch (RX.State) {  		 		
  	case RX_IDLE:
  		// se verifica que RB4 este en 1 para comenzar a medir LAMBDA
  		if (input(RX_IN) == 1) {
  			// habilita la interrupcion por TMR0
  			set_timer0 (0x00);
  			clear_interrupt (INT_TIMER0);
  			enable_interrupts (INT_TIMER0);
  			// inicializa el valor del tiempo en bajo, en alto y lambda
  			Lambdatime = 0x0000;
  			PulseHighTime = 0x0000;
  			PulseLowTime = 0x0000;
  			// borra el contador del TMR1
  			set_timer1(0x0000);
  			// setea el siguiente paso
  			RX.State = RX_GET_LAMBDA;
  		}
  	break;
  	
  	case RX_GET_LAMBDA:
  		// se verifica que RB4 este en 0 para terminar a medir LAMBDA
  		if (input(RX_IN) == 0) {
  			// resetea el TMR0
  			set_timer0 (0x00);
	  		// carga el valor de lambda
	  		Lambdatime = get_timer1();
	  		// borra el contador del TMR1
	  		set_timer1 (0x0000);
	  		// borra los contadores de longitud de pulso
	  		PulseHighTime = 0x0000;
	  		PulseLowTime = 0x0000;
	  		// borra el contador de bit
	  		RX.CounterBits = 0x00;
	  		// setea el siguiente paso
	  		RX.State = RX_GET_PACKAGE;
	  	}	  		  				
  	break;
  	
  	case RX_GET_PACKAGE:  			  						
			// se verifica el estado del RB4 para ver el flanco
			if (input(RX_IN) == 1) {
				// resetea el TMR0
  			set_timer0 (0x00);
				// carga el valor del pulso en bajo
				PulseLowTime = get_timer1();
				// borra el contador del TMR1
				set_timer1 (0x0000);  				  				  			
			}
			else {
				// resetea el TMR0
  			set_timer0 (0x00);
				// carga el valor del pulso en alto
				PulseHighTime = get_timer1();
				// borra el contador del TMR1
				set_timer1 (0x0000);
			}

			// se verifica si ya se levanto un bit completo
			if ((PulseHighTime > 0) && (PulseLowTime > 0)) {
				// se analiza el resultado
				if (PulseLowTime > ((2 * Lambdatime) + LAMBDA_ERROR)){
					// deshabilita todas las interrupciones
					disable_interrupts(GLOBAL);
					// seta el resultado de la lectura
					RX.Result = RX_GET_ERROR;  					  				
				}
				else if (PulseLowTime < (Lambdatime - LAMBDA_ERROR)) {
					// deshabilita todas las interrupciones
					disable_interrupts(GLOBAL);
					// seta el resultado de la lectura
					RX.Result = RX_GET_ERROR;
				}
				else if (PulseHighTime > ((2 * Lambdatime) + LAMBDA_ERROR)) {
					// deshabilita todas las interrupciones
					disable_interrupts(GLOBAL);
					// seta el resultado de la lectura
					RX.Result = RX_GET_ERROR;	
				}
				else if (PulseHighTime < (Lambdatime - LAMBDA_ERROR)) {
					// deshabilita todas las interrupciones
					disable_interrupts(GLOBAL);
					// seta el resultado de la lectura
					RX.Result = RX_GET_ERROR;
				}
				else {
					RX.CounterBits++;
					RX.Package <<= 1;	  				
					RX.Package &= 0xFFFFFFFE;
					if (PulseLowTime > PulseHighTime)
						RX.Package |= 1;
					// seta el resultado de la lectura
					RX.Result = RX_GET_OK; 				
  			}
  			// reseta los medidores de pulso
  			PulseLowTime = 0x0000;
  			PulseHighTime = 0x0000; 			
  		}  			  						  			
  	break;
  	
  	case RX_GET_PACKAGE_OK:
  	break;
  	
  	case RX_GET_PACKAGE_ERROR:
  	break;
  		  		
  	default:
  		// deshabilita todas las interrupciones
			disable_interrupts(GLOBAL);
  		// seta el resultado de la lectura
  		RX.Result = RX_GET_ERROR;
  	break;
  }
  // reseta el flag
  clear_interrupt(INT_RB);
}

// interrupcion por overflow en el TMR0
#INT_TIMER0
void interrupt_Timer0 () {
	
	// reseta el flag
  clear_interrupt (INT_TIMER0);	
	// deshabilita todas las interrupciones
	disable_interrupts(GLOBAL);	   
  // setea el siguiente estado
  RX.State = RX_GET_PACKAGE_ERROR;
}

// funcion principal
void main() {

	unsigned char i;
	unsigned long long long_buf_address1, long_buf_address2;
	unsigned char chr_buf_data1, chr_buf_data2;
	char buffer_str_lcd[20];
	
	// deshabilita todas las interrupciones
	disable_interrupts(GLOBAL);
	
  // setea el clock interno
  setup_oscillator(OSC_4MHZ);
  
  // setea el puerto B
  set_tris_B (0x14);												// RB4 = 1, RB2 = 1
  port_b_pullups (TRUE);
	Lcd_Backlight_Off;
	Buzzer_Off;
	
	// setea el timer 0
	setup_timer_0 (RTCC_DIV_4 | RTCC_INTERNAL);
	
	// setea el timer 1			
  setup_timer_1 (T1_DIV_BY_1 | T1_INTERNAL);
	
	// deshabilita el comparador
	setup_ccp1 (CCP_OFF);
	
  // habilita interrupcion por cambio en RB4 - RB7 
	enable_interrupts (INT_RB);
	
	// se enciende el BUZZER por un instante 
	Buzzer_On;
	delay_ms(200);
	Buzzer_Off;
	
	// se inicializa el LCD
	Lcd_Backlight_Off;
	Lcd_Init();
	Lcd_PutC('\f');
	Lcd_Backlight_On;	
	
	// se muestra el mensaje de inicio de la demo
	sprintf(buffer_str_lcd, "%s", "Cika Electronica");
	Lcd_OutStr(1, 1, buffer_str_lcd);
	sprintf(buffer_str_lcd, "%s", "Demo RX HT6P20");
	Lcd_OutStr(2, 2, buffer_str_lcd);
	delay_ms(2000);			
		
	// se muestra el mensaje de espera en el display
	Lcd_PutC('\f');
	sprintf(buffer_str_lcd, "%s", "Esperando");
	Lcd_OutStr(5, 1, buffer_str_lcd);
	sprintf(buffer_str_lcd, "%s", "Recepcion...");
	Lcd_OutStr(3, 2, buffer_str_lcd);	
	
	// inicializa la structura de recepcion
	RX.Model = HT6P20B;
	RX.Package = 0;
	RX.Address = 0;
	RX.strAddress[0] = 0; 
	RX.Data = 0;
	RX.strData[0] = 0;
	RX.State = RX_IDLE;
	RX.Result = RX_GET_OK;
	RX.CounterBits = 0;
	
	// habilita las interrupciones globales
	enable_interrupts(GLOBAL);
	
	// loop infinito
	for ( ; ; ) {
		switch (RX.State) {		
			case RX_IDLE:				
				// verifica el modelo de encoder a decodificar
				if (input(RX_MODEL) == 0)
					RX.Model = HT6P20B;
				else
					RX.Model = HT6P20D;				
			break;
			
			case RX_GET_LAMBDA:
				RX.State = RX_GET_LAMBDA;
			break;
			
			case RX_GET_PACKAGE:
				if (RX.State == RX_GET_PACKAGE) {
					// se verifica si ocurrio algun error en la recepcion
					if (RX.Result == RX_GET_OK) {
						// se verifica si ya se recibieron todos los bits
						if (RX.CounterBits == HT6P20X2_LEN_PKG) {
							// deshabilita las interrupciones globales
							disable_interrupts(GLOBAL);
							// deshabilita la interrupcion por TMR0
  						disable_interrupts (INT_TIMER0);					
							// resetea el contador de bits
							RX.CounterBits = 0;
							// se verifica el ENDCODE
							if ((char)(RX.Package & 0x0000000F) != END_CODE) {
								// seta el siguiente estado
								RX.State = RX_GET_PACKAGE_ERROR;
							}
							else {
	  						// seta el siguiente estado
								RX.State = RX_GET_PACKAGE_OK;																
							}
						}																							
					}
					else {
						// deshabilita las interrupciones globales
						disable_interrupts(GLOBAL);
						// deshabilita la interrupcion por TMR0
						clear_interrupt (INT_TIMER0);
  					disable_interrupts (INT_TIMER0);
						// seta el siguiente estado
						RX.State = RX_GET_PACKAGE_ERROR;	
					}
				}			
			break;
			
			case RX_GET_PACKAGE_OK:
				// se parsean los datos
				if (RX.Model == HT6P20B) {
					// se rotan los bits de la direccion para dejar el MSB primero
					long_buf_address1 = RX.Package >> (HT6P20X2_LEN_PKG - HT6P20B_LEN_ADDRESS);
					long_buf_address1 &= 0x003FFFFF;
					long_buf_address2 = 0;
					long_buf_address2 = long_buf_address2 | (long_buf_address1 & 0x00000001);
	        for (i = 0; i < (HT6P20B_LEN_ADDRESS - 1); i++){
						long_buf_address2 <<= 1;
						long_buf_address1 >>= 1;
						long_buf_address2 = long_buf_address2 | (long_buf_address1 & 0x00000001);						
					}
					RX.Address = long_buf_address2;
					
					// se rotan los bits de datos para dejar el MSB primero
					chr_buf_data1 = (char)(RX.Package >> 4);
					chr_buf_data1 &= 0x03;
					chr_buf_data2 = 0;
					chr_buf_data2 = chr_buf_data2 | (chr_buf_data1 & 0x01);
					chr_buf_data2 <<= 1;
					chr_buf_data1 >>= 1;
					chr_buf_data2 = chr_buf_data2 | (chr_buf_data1 & 0x01);
					RX.Data = chr_buf_data2;												
				}
				else {
					// se rotan los bits de la direccion para dejar el MSB primero
					long_buf_address1 = RX.Package >> (HT6P20X2_LEN_PKG - HT6P20D_LEN_ADDRESS);
					long_buf_address1 &= 0x000FFFFF;															
					long_buf_address2 = 0;
					long_buf_address2 = long_buf_address2 | (long_buf_address1 & 0x00000001);
											        
					for (i = 0; i < (HT6P20D_LEN_ADDRESS - 1); i++){
						long_buf_address2 <<= 1;
						long_buf_address1 >>= 1;
						long_buf_address2 = long_buf_address2 | (long_buf_address1 & 0x00000001);						
					}
					RX.Address = long_buf_address2;
					
					// se rotan los bits de datos para dejar el MSB primero
					chr_buf_data1 = (char)(RX.Package >> 4);
					chr_buf_data1 &= 0x0F;
					chr_buf_data2 = 0;
					chr_buf_data2 = chr_buf_data2 | (chr_buf_data1 & 0x01);
					for(i = 0; i < 3; i++){
						chr_buf_data2 <<= 1;
						chr_buf_data1 >>= 1;
						chr_buf_data2 = chr_buf_data2 | (chr_buf_data1 & 0x01);
					}
					RX.Data = chr_buf_data2;										
				}
				
				// se convierte el dato recibido en ASCII para el display
				LongToStrHex(RX.Address, RX.strAddress);
				ByteToStrHex(RX.Data, RX.strData);
				
				// se indica con el buzzer que los datos llegaron OK
				Buzzer_On;
				delay_ms(200);
				Buzzer_Off;
				
				// se muestran los datos recibidos en el display
				Lcd_PutC('\f');
				sprintf(buffer_str_lcd, "Dir.:0x%s", RX.strAddress);
				Lcd_OutStr(1, 1, buffer_str_lcd);
				sprintf(buffer_str_lcd, "Dato:0x%s", RX.strData);
				Lcd_OutStr(1, 2, buffer_str_lcd);
				delay_ms(500);
			
			case RX_GET_PACKAGE_ERROR:
				// deshabilita la interrupcion por TMR0
  			disable_interrupts (INT_TIMER0);
  			
			default: 
				RX.Package = 0;
				RX.Address = 0;
				RX.strAddress[0] = 0; 
				RX.Data = 0;
				RX.strData[0] = 0;
				RX.State = RX_IDLE;
				RX.Result = RX_GET_OK;
				RX.CounterBits = 0;
				
				// reseta el flags
  			clear_interrupt(INT_RB);
  			clear_interrupt (INT_TIMER0);
				
				// habilito las interrupciones globales
				enable_interrupts(GLOBAL);
			break;
		}
	}
}